#include <FM94Elem.h>
#include <string.h>
#include <math.h>

/* 04/02/2014 - corrected cooefficients for fast deep and T-11.  Changed a coefficient for T-11 to 1.7790 from 1.7779
                Changed b coefficient fast deep to 6.346 from 6.390.
*/

// 11/5/14 PIC

FM94Elem::FM94Elem()
	{
	classid = 0;
	version = 0;
	offset = 0;
	minimum = 0;
	bits = 0;
	maximum = 0;
	scale = 0;
	Desc[0] = '\0';
	units[0] = '\0';
	Valid = false;
	for (int i=0; i<sizeof(Value); i++)
		{Value.StrVal[i]=(char)255;}
	}

FM94Elem::~FM94Elem()
{}


void FM94Elem::Init(int id, int ver, int off, int min, int inbits, int max, int inscale, char *indesc, char *inunits)
	{
	classid = id;
	version = ver;
	offset = off;
	minimum = min;
	bits = inbits;
	maximum = max;
	scale = inscale;
	strcpy(Desc,indesc);
	strcpy(units, inunits);
	Valid = false;
	for (int i=0; i<sizeof(Value); i++)
		{Value.StrVal[i]=(char)255;}

	}

void FM94Elem::SetValue(char *valuein)
{
	for (int i=0; i<sizeof(Value); i++)
		{Value.StrVal[i]=(char)255;}

	int k = strlen(valuein);
	if (k>60)
	{k=60;}
	int k1 = GetMaxBits()/8;


	Valid = false;

	strncpy(Value.StrVal, valuein, GetMaxBits()/8);
	Valid = true;
}

void FM94Elem::SetValue(int valuein)
{	
	for (int i=0; i<sizeof(Value); i++)
		{Value.StrVal[i]=(char)255;}

	if(strcmp(units,"CCITT IA5"))
	{
		int x = GetMinimum();
		int y = GetMaximum();
		if((valuein>= GetMinimum())&&(valuein<=GetMaximum()))
		{

			Value.Integ = valuein;
			Valid = true;
		}
		else
		{
			Value.Integ = -1;
			Valid = false;
		}

	}
}


void FM94Elem::SetValue(unsigned long valuein)
{	
	for (int i=0; i<sizeof(Value); i++)
		{Value.StrVal[i]=(char)255;}

	if(strcmp(units,"CCITT IA5"))
	{
		int x = GetMinimum();
		int y = GetMaximum();
		if((valuein>= GetMinimum())&&(valuein<=GetMaximum()))
		{

			Value.Integ = valuein;
			Valid = true;
		}
		else
		{
			Value.Integ = -1;
			Valid = false;
		}

	}
}

bool FM94Elem::IsValid()
{return Valid;}

int FM94Elem::GetClassId()
	{return classid;}
int FM94Elem::GetVersion()
	{return version;}
int FM94Elem::GetOffset()
	{return offset;}
int FM94Elem::GetMinimum()
	{return minimum;}
int FM94Elem::GetMaxBits()
	{return bits;}
int FM94Elem::GetMaximum()
	{return maximum;}
int FM94Elem::GetScale()
	{return scale;}
char * FM94Elem::GetDesc()
	{//	char *retval[sizeof(Desc)];	
	//	strcpy(retval,Desc);
	//	return retval;
	return Desc;}
char * FM94Elem::GetUnits()
	{return units;}

FM94Value FM94Elem::GetValue()
{return Value;}

char *FM94Elem::GetBufrVal()
{
	for (int i=0; i<60; i++)
	{
		BufrVal.StrVal[i]='\0';
	}

	int numbytes = GetMaxBits()/8;
	if(GetMaxBits()%8 != 0){numbytes+=1;}

	if(strcmp(GetUnits(),"CCITT IA5") == 0) 
	{
		char *A;
		A= GetValue().StrVal;
		strncpy(BufrVal.StrVal,GetValue().StrVal, numbytes);
		for(int i = numbytes;i=0; i--)
		{
			BufrVal.StrVal[numbytes]='\0';
			if(BufrVal.StrVal[i]==-1)
			{
				BufrVal.StrVal[i]='\0';
			}
			else
			{
				break;
			}
		}
	}
	else
	{
		unsigned x = (unsigned) GetValue().Integ;
		int NoShifts = 32 - GetMaxBits();
		 x= x * (unsigned) (pow(2.0,double(NoShifts)));
		 BufrVal.Integ = x;

		 // Swap the bytes to make integer appear as bit stream (little endian)

		 char a = BufrVal.StrVal[0];
		 BufrVal.StrVal[0] = BufrVal.StrVal[3];
		 BufrVal.StrVal[3] = a;
		 a = BufrVal.StrVal[1];
		 BufrVal.StrVal[1] = BufrVal.StrVal[2];
		 BufrVal.StrVal[2] = a; 
	}

	return BufrVal.StrVal;
}

	

// FXY1011

// Body
// FXY1011 Ship Call Sign
FXY1011::FXY1011()
{
	FM94Elem::Init(1011,1,0,0,72,9,1,"SHIP CALL SIGN","CCITT IA5");
}

//FXY1012 Ship's Direction
FXY1012::FXY1012()
{
	Init(1012,1,0,0,9,510,0,"DIRECTION SHIP IS HEADING","DEGREES TRUE");
}

//FXY1013 Ship's Speed
FXY1013::FXY1013()
{
	Init(1013,1,0,0,13,8190,2,"SHIP'S SPEED","METERS/SEC");
}

//FXY1013A Ship's Speed
FXY1013A::FXY1013A()
{
	Init(1013,1,0,0,10,1022,0,"SHIP'S SPEED","METERS/SEC");
}

// FXY1036 Agency in charge of operating the observing platform
FXY1036::FXY1036()
{
	
	AgencyTable A[ Elem1036] = 
	{
		{36001, "Australia, Bureau of Meteorology (BoM)"},
		{36002, "Australia, Joint Australian Facility for Ocean Observing Systems (JAFOOS)"},
		{36003, "Australia, the Commonwealth Scientific and Industrial Research Organisation (CSIRO)"},
		{124001, "Canada, Marine Environmental Data Service (MEDS)"},
		{124002, "Canada, Institute of Ocean Sciences (IOS)"},
		{156001, "China, The State Oceanic Administration"},
		{156002, "China, Second Institute of Oceanography, State Oceanic Administration"},
		{156003, "China, Institute of Ocean Technology"},
		{250001, "France, Institut de Recherche pour le Dveloppement (IRD)"},
		{250002, "France, Institut Franais de Recherche pour l'Exploitation de la mer (IFREMER)"},
		{276001, "Germany, Bundesamt fuer Seeschiffahrt und Hydrographie (BSH)"},
		{276002, "Germany, Institut fuer Meereskunde, Kiel"},
		{356001, "India, National Institute of Oceanography (NIO)"},
		{356002, "India, National Institute for Ocean Technology (NIOT)"},
		{356003, "India, National Centre for Ocean Information Service"},
		{392001, "Japan, Japan Meteorological Agency (JMA)"},
		{392002, "Japan, Frontier Observational Research System for Global Change"},
		{392003, "Japan, Japan Marine Science and Technology Centre (JAMSTEC)"},
		{410001, "Republic of Korea, Seoul National University"},
		{410002, "Republic of Korea, Korea Ocean Research and Development Institute  (KORDI)"},
		{410003, "Republic of Korea, Meteorological Research Institute"},
		{540001, "New Caledonia, Institut de Recherche pour le Dveloppement (IRD)"},
		{554001, "New Zealand, National Institute of Water and Atmospheric Research (NIWA)"},
		{643001, "Russian Federation, State Oceanographic Institute of Roshydromet"},
		{643002, "Russian Federation, Federal Service for Hydrometeorology and Environmental Monitoring"},
		{724001, "Spain, Instituto Espaol de Oceanografia"},
		{826001, "United Kingdom, Hydrographic Office"},
		{826002, "United Kingdom, Southampton Oceanography Centre (SOC)"},
		{840001, "USA, NOAA Atlantic Oceanographic and Meteorological Laboratories (AOML)"},
		{840002, "USA, NOAA Pacific Marine Environmental Laboratories (PMEL)"},
		{840003, "USA, Scripps Institution of Oceanography (SIO)"},
		{840004, "USA, Woods Hole Oceanographic Institution (WHOI)"},
		{840005, "USA, University of Washington"},
		{840006, "USA, Naval Oceanographic Office"},
		{1048575, "Missing value"}
	};
	for (int i = 0; i< Elem1036; i++)
	{
		Table.push_back(A[i]);
	}


	Init(1036,1,0,0,20,1048576,0,"Agency in charge of operating the observing platform","CODE TABLE");
};

//FXY1080  Ship Line Number According to SOOP
FXY1080::FXY1080()
{
	FM94Elem::Init(1080,1,0,0,48,6,1,"SOOP LINE NUMBER","CCITT IA5");
}
//FXY1200  Ship Name
FXY1200::FXY1200()
{
	Init(1200,1,0,0,240,30,1,"SHIP NAME","CCITT IA5");
}

// Ship's Lloyds (IMO) number
FXY1201::FXY1201()
{
	Init(1201,1,0,0,24,9999999,1,"IMO NUMBER","NUMERIC");
}

FXY2002::FXY2002()
{
	CodeTable A[Elem2002] =
	{
		{0,"Uncertified Instruments meters/sec"},
		{1,"Certified Instruments meters/sec"},
		{2,"Uncertified Instruments knots"},
		{3,"Certified Instruments knots"},
		{4,"Uncertified Instruments kilometers/hour"},
		{5,"Certified Instruments kilometers/hour"}
	};

	for (int i = 0; i< Elem2002; i++)
	{
		Table.push_back(A[i]);
	}


	Init(2002,1,0,0,4,14,0,"TYPE OF INSTR. FOR WIND MEASUREMENT","FLAG TABLE");
}

int FXY2002::SetValue(int val)
{

	for (int i=0;i < Elem2002;i++)
	{
		if(Table.at(i).Code==val)
		{
			Value.Integ = val;
			Valid = true;
			return 0;
		}
	}

	Value.Integ = -1;
	Valid = false;
	return 0;
}	

// FXY2030  Method of current measurement
FXY2030::FXY2030()
{
	CodeTable A[Elem2030] =
	{
		{0,"Reserved"},
		{1,"ADCP Acoustic Doppler Current Profiler"},
		{2,"GEK Geomagnetic ElectroKinetograph"},
		{3,"Ship's set and drift determined by fixes 3-6 hours apart"},
		{4,"Ship's set and drift determined by fixes > 6 hours but < 12 hours apart"},
		{5,"Drift of buoy"},
		{6,"Reserved"}
	
	};

	for (int i = 0; i<Elem2030; i++)
	{
		Table.push_back(A[i]);
	}


	Init(2030,1,0,0,3,6,0,"METHOD OF CURRENT MEASUREMENT","CODE TABLE");
}



// FXY2038 METH OF SEA TEMPERATURE MEASUREMENT
FXY2038::FXY2038()
{
	Init(2038,1,0,0,3,6,0,"METH OF SEA TEMPERATURE MEASUREMENT","NUMERIC");
}

// FXY2039 METH OF WET BULB TEMPERATURE MEASUREMENT
FXY2039::FXY2039()
{
	Init(2039,1,0,0,3,6,0,"METH OF WET BULB TEMPERATURE MEASUREMENT","NUMERIC");
}

// FXY2171 Instrument serial number for water temperature profile measurement
FXY2171::FXY2171()
{
	Init(2171,1,0,0,64,0,0,"INSTRUMENT SERIAL NUMBER FOR WATER TEMPERATURE PROFILE MEASUREMENT" ,"CCITT IA5");
}


// FXY4001 Year
FXY4001::FXY4001()
{
	Init(4001,1,0,0,12,4094,0,"YEAR","NUMERIC");
}

// FXY4002 Month
FXY4002::FXY4002()
{
	Init(4002,1,0,1,4,12,0,"MONTH OF YEAR","NUMERIC");
}

// FXY4003 Day of Month
FXY4003::FXY4003()
{
	Init(4003,1,0,1,6,31,0,"DAY OF MONTH","NUMERIC");
}

// FXY4004 Hours
FXY4004::FXY4004()
{
	Init(4004,1,0,0,5,23,0,"HOUR OF DAY","NUMERIC");
}

// FXY4005 Minutes
FXY4005::FXY4005()
{
	Init(4005,1,0,0,6,59,0,"MINUTE OF HOUR","NUMERIC");
}

// FXY4006 Seconds
FXY4006::FXY4006()
{
	Init(4006,1,0,0,6,59,0,"SECOND OF MINUTE","NUMERIC");
}

// FXY4025 Time period or displacement Minutes
FXY4025::FXY4025()
{
	Init(4025,1,-2048,0,12,4094,0,"MINUTES OF DISPLACEMENT","NUMERIC");
}

// FXY4025A Time period or displacement Minutes
FXY4025A::FXY4025A()
{
	Init(4025,1,0,0,16,65534,0,"MINUTES OF DISPLACEMENT","NUMERIC");
}

// FXY4209 MINUTE OF MONTH
FXY4209::FXY4209()
{
	Init(4209,1,0,0,16,44640,0,"MINUTE OF MONTH","NUMERIC");
}

// FXY5001 Latitude High Accuracy
FXY5001::FXY5001()
{
	Init(5001,1,-9000000,0,25,18000000,5,"LATITUDE(HIGH ACCURACY)","DEGREES");
}

// FXY5002 Latitude Coarse Accuracy
FXY5002::FXY5002()
{
	Init(5002,1,-9000, 0,15,18000,2,"LATITUDE(COARSE ACCURACY)","DEGREES");
}

// FXY5011 Latitude Increment High Accuracy
FXY5011::FXY5011()
{
	Init(5011,1,-9000000,0,25,18000000,5,"LATITUDE INCREMENT (HIGH ACCURACY)","DEGREES");
}

// FXY5012 Latitude Increment Coarse Accuracy
FXY5012::FXY5012()
{
	Init(5012,1,-9000,0,15,18000,2,"LATITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY5012S4 Latitude Increment Coarse Accuracy SEAS IV
FXY5012S4::FXY5012S4()
{
	Init(5012,1,-30,0,6,62,1,"LATITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY5012S4A Latitude Increment Coarse Accuracy SEAS IV AMVER SP
FXY5012S4A::FXY5012S4A()
{
	Init(5012,1,-4000,0,13,8190,2,"LATITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY5036 Ship Transect Number
FXY5036::FXY5036()
{
	Init(5036,1,0,0,7,126,0,"SHIP TRANSECT NUMBER","NUMERIC");
}

// FXY6001 Longitude High Accuracy
FXY6001::FXY6001()
{
	Init(6001,1,-18000000,0,26,36000000,5,"LONGITUDE(HIGH ACCURACY)","DEGREES");
}

// FXY6002 Longitude Coarse Accuracy
FXY6002::FXY6002()
{
	Init(6002,1,-18000,0,16,36000,2,"LONGITUDE(COARSE ACCURACY)","DEGREES");
}

// FXY6011 Longitude Increment High Accuracy
FXY6011::FXY6011()
{
	Init(6011,1,-18000000,0,26,36000000,5,"LONGITUDE INCREMENT (HIGH ACCURACY)","DEGREES");
}

// FXY6012 Longitude Increment Coarse Accuracy
FXY6012::FXY6012()
{
	Init(6012,1,-18000,0,16,36000,2,"LONGITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY6012 Longitude Increment Coarse Accuracy SEAS IV
FXY6012S4::FXY6012S4()
{
	Init(6012,1,-60,0,7,126,1,"LONGITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY6012 Longitude Increment Coarse Accuracy SEAS IV AMVER SP
FXY6012S4A::FXY6012S4A()
{
	Init(6012,1,-8000,0,14,16382,1,"LONGITUDE INCREMENT (COARSE ACCURACY)","DEGREES");
}

// FXY7062 Depth Below Sea/Water Surface

FXY7062::FXY7062()
{
	Init(7062,1,0,0,17,131070,1,"DEPTH BELOW SEA SURFACE","METERS");
}

// FXY7062S Depth Below Sea/Water Surface

FXY7062S::FXY7062S()
{
	Init(7062,1,0,0,14,16382,2,"DEPTH BELOW SEA SURFACE","METERS");
}

//FXY10051 PRESSURE REDUCED TO MSL
FXY10051::FXY10051()
{
	Init(10051,1,8600,0,12,4094,-1,"PRESSURE REDUCED TO MSL","PASCALS");
}

//FXY10061  3 HOUR PRESSURE CHANGE
FXY10061::FXY10061()
{
	Init(10061,1,-500,0,10,1000,-1,"3 HOUR PRESSURE CHANGE","PASCALS");
}

// FXY10063  CHARACTERISTIC OF PRESSURE TENDENCY
FXY10063::FXY10063()
{
	Init(10063,1,0,0,4,14,1,"CHARACTERISTIC OF PRESSURE TENDENCY","NUMERIC");
}


//FXY11001 Wind Direction
FXY11001::FXY11001()
{
	Init(11001,1,0,0,9,510,0,"WIND DIRECTION","DEGREES TRUE");
}

//FXY11001A Wind Direction
FXY11001A::FXY11001A()
{
	Init(11001,1,0,0,6,36,0,"WIND DIRECTION","DEGREES TRUE");
}

//FXY11002 Wind Speed
FXY11002::FXY11002()
{
	Init(11002,1,0,0,12,4094,1,"WIND SPEED","METERS/SEC");
}

//FXY11002 Wind Speed
FXY11223::FXY11223()
{
	Init(11223,1,0,0,8,200,1,"WIND SPEED","KNOTS");
}

//FXY12004 Dry Bulb Temperature Degrees K = 273.15+ Degrees C
FXY12004::FXY12004()
{
	Init(12004,1,0,0,12,4094,1,"DRY BULB TEMPERATURE","DEGREES KELVIN");
}

//FXY12004A Dry Bulb Temperature Degrees K = 273.15+ Degrees C
FXY12004A::FXY12004A()
{
	Init(12004,1,1802,0,11,2046,1,"DRY BULB TEMPERATURE","DEGREES KELVIN");
}

//FXY12005A Wet Bulb Temperature Degrees K = 273.15+ Degrees C
FXY12005A::FXY12005A()
{
	Init(12005,1,1802,0,11,2046,1,"WET BULB TEMPERATURE","DEGREES KELVIN");
}

//FXY20003 PRESENT WEATHER
FXY20003::FXY20003()
{
	Init(20003,1,0,0,7,126,0,"PRESENT WEATHER","NUMERIC");
}

//FXY20004 PAST WEATHER (1)
FXY20004::FXY20004()
{
	Init(20004,1,0,0,4,14,0,"PAST WEATHER (1)","NUMERIC");
}

//FXY20004 PAST WEATHER (1)
FXY20005::FXY20005()
{
	Init(20005,1,0,0,4,14,0,"PAST WEATHER (2)","NUMERIC");
}

//FXY20011 CLOUD AMOUNT
FXY20011::FXY20011()
{
	Init(20011,1,0,0,4,14,0,"CLOUD AMOUNT","NUMERIC");
}

//FXY20012 CLOUD TYPE
FXY20012::FXY20012()
{
	Init(20012,1,0,0,4,14,0,"CLOUD TYPE","NUMERIC");
}

//FXY20031 THICKNESS OF ICE ACCRETION
FXY20031::FXY20031()
{
	Init(20031,1,0,0,7,126,2,"THICKNESS OF ICE ACCRETION","METERS");
}

//FXY20032 RATE OF ICE ACCRETION
FXY20032::FXY20032()
{
	Init(20032,1,0,0,3,6,0,"RATE OF ICE ACCRETION","NUMERIC");
}

//FXY20033 CAUSE OF ICE ACCRETION
FXY20033::FXY20033()
{
	Init(20033,1,0,0,3,6,0,"CAUSE OF ICE ACCRETION","NUMERIC");
}

//FXY20034 SEA ICE CONCENTRATION
FXY20034::FXY20034()
{
	Init(20034,1,0,0,4,14,0,"CAUSE OF ICE ACCRETION","NUMERIC");
}

//FXY20035 AMOUNT AND TYPE OF ICE
FXY20035::FXY20035()
{
	Init(20035,1,0,0,4,14,0,"AMOUNT AND TYPE OF ICE","NUMERIC");
}

//FXY20036 ICE SITUATION
FXY20036::FXY20036()
{
	Init(20036,1,0,0,4,14,0,"ICE SITUATION","NUMERIC");
}

//FXY20037 ICE DEVELOPMENT
FXY20037::FXY20037()
{
	Init(20037,1,0,0,4,14,0,"ICE DEVELOPMENT","NUMERIC");
}

//FXY20193 LOWEST CLOUDS 
FXY20193::FXY20193()
{
	Init(20193,1,0,0,4,14,0,"LOWEST CLOUDS","NUMERIC");
}

//FXY20194 HORIZONTAL VISIBILITY
FXY20194::FXY20194()
{
	Init(20194,1,0,0,4,14,0,"HORIZONTAL VISIBILITY","NUMERIC");
}

//FXY20195 CLOUD COVERAGE TENTHS
FXY20195::FXY20195()
{
	Init(20195,1,0,0,4,14,0,"CLOUD COVERAGE TENTHS","NUMERIC");
}



//FXY20197 BEARING OF SEAS ICE
FXY20197::FXY20197()
{
	Init(20197,1,0,0,4,14,0,"BEARING OF SEAS ICE","NUMERIC");
}

//FXY20201 MEASURED WIND INDICATOR
FXY20201::FXY20201()
{
	Init(20201,1,0,0,2,1,0,"MEASURED WIND INDICATOR","NUMERIC");
}

//FXY22003 Direction of SWELL WAVES
FXY22003::FXY22003()
{
	Init(22003,1,0,0,6,62,0,"DIRECTION OF SWELL WAVES","DEGREES TRUE");
}

//FXY22012 PERIOD OF WIND WAVES
FXY22012::FXY22012()
{
	Init(22012,1,0,0,4,14,0,"PERIOD OF WIND WAVES","SECONDS");
}


//FXY22013 PERIOD OF SWELL WAVES
FXY22013::FXY22013()
{
	Init(22013,1,0,0,5,30,0,"PERIOD OF SWELL WAVES","SECONDS");
}

//FXY22004 Direction of Current
FXY22004::FXY22004()
{
	Init(22004,1,0,0,9,510,0,"DIRECTION OF CURRENT","DEGREES TRUE");
}

//FXY22031 Speed of Current
FXY22031::FXY22031()
{
	Init(22031,1,0,0,13,8190,2,"SPEED OF CURRENT","METERS/SEC");
}


//FXY22042 Sea/Water Temperature Degrees K = 273.15+ Degrees C
FXY22042::FXY22042()
{
	Init(22042,1,0,0,12,4094,1,"SEA TEMPERATURE","DEGREES KELVIN");
}

//FXY22042S Sea/Water Temperature Degrees C 
FXY22042S::FXY22042S()
{
	Init(22042,1,-400,0,12,4094,2,"SEA TEMPERATURE","DEGREES CELSIUS");
}

//FXY22042T Sea/Water Temperature Degrees C 
FXY22042T::FXY22042T()
{
	Init(22042,1,-4000,0,16,65534,3,"SEA TEMPERATURE","DEGREES CELSIUS"); // used for SEABIRD TSG
}

//FXY22042S4 Sea/Water Temperature Degrees K 
FXY22042S4::FXY22042S4()
{
	Init(22042,1,2682,0,9,510,1,"SEA TEMPERATURE","DEGREES KELVIN"); // used for SEAIV Weather Sea Temp
}




//FXY22044 Sound Velocity
FXY22044::FXY22044()
{
	Init(22044,1,0,0,14,16382,2,"SOUND VELOCITY","METERS PER SECOND");
}

//FXY22062 Salinity parts per thousand
FXY22062::FXY22062()
{
	Init(22062,1,0,0,14,16382,1,"SALINITY","PART PER THOUSAND");
}

//FXY22063 Total Water Depth
FXY22063::FXY22063()
{
	Init(22063,1,0,0,14,16382,0,"TOTAL WATER DEPTH","METERS");

}
//FXY22062 Salinity parts per thousand
FXY22066::FXY22066()
{
	Init(22066,1,0,0,14,16382,1,"CONDUCTIVITY","SIEMENS PER METER");
}

//FXY22067    Instrument type for water temperature profile measurement

FXY22067::FXY22067()
{
	Profiler A[sizeof(Instr)/sizeof(Instr[0])] =
	{
		{1,  6.472, -2.16,	460,	"Sippican T-4"},
		{2,  6.691, -2.25,	460,	"Sippican T-4"},
		{11, 6.828, -1.82,	1830,	"Sippican T-5"},
		{21, 6.346, -1.82,	1000,	"Sippican Fast Deep"},
		{31, 6.472, -2.16,	460,	"Sippican T-6"},
		{32, 6.691, -2.25,	460,	"Sippican T-6"},
		{41, 6.472, -2.16,	760,	"Sippican T-7"},
		{42, 6.691, -2.25,	760,	"Sippican T-7"},
		{51, 6.472, -2.16,	760,	"Sippican Deep Blue"},
		{52, 6.691, -2.25,	760,	"Sippican Deep Blue"},
		{61, 6.301, -2.16,	200,	"Sippican T-10"},
		{71, 1.779, -0.255, 460,	"Sippican T-11"},
		{201,6.472, -2.16,	460,	"TSK T-4"},
		{202,6.691, -2.25,	460,	"TSK T-4"},
		{211,6.472, -2.16,	460,	"TSK T-6"},
		{212,6.691, -2.25,	460,	"TSK T-6"},
		{221,6.472, -2.16,	760,	"TSK T-7"},
		{222,6.691, -2.25,	760,	"TSK T-7"},
		{231,6.828, -1.82,	1830,	"TSK T-5"},
		{241,6.301, -2.16,	200,	"TSK T-10"},
		{251,6.472, -2.16,	760,	"TSK Deep Blue"},
		{252,6.691, -2.25,	760,	"TSK Deep Blue"},
		{261, -1,    -1,	-1,		"TSK AXBT"},
		{401,6.301, -2.16,	-1,		"Sparton XBT-1"},
		{411,5.861, -0.0904,-1,		"Sparton XBT-3"},
		{421,6.472, -2.16,	-1,		"Sparton XBT-4"},
		{431,6.828, -1.82,	-1,		"Sparton XBT-5"},
		{441,6.828, -1.82,	-1,		"Sparton XBT-5DB"},
		{451,6.472, -2.16,	-1,		"Sparton XBT-6"},
		{461,6.472, -2.16,	-1,		"Sparton XBT-7"},
		{462,6.705, -2.28,	-1,		"Sparton XBT-7"},
		{471,6.472, -2.16,	-1,		"Sparton XBT-7DB"},
		{481,6.301, -2.16,	-1,		"Sparton XBT-10"},
		{491,6.472, -2.16,	-1,		"Sparton XBT-20"},
		{501,6.472, -2.16,	-1,		"Sparton XBT-20DB"},
		{510,1.524,  0,		-1,		"Sparton 536 AXBT-7"},
		{700, -1,	-1,		-1,		"Sippican XCTD Standard"},
		{730, -1,	-1,		-1,		"Sippican SXCTD"},
		{741, -1,	-1,		-1,		"TSK XCTD"},
		{751, -1,	-1,		-1,		"TSK AXCTD"},
		{800, -1,	-1,		-1,		"Mechanical BT"},
		{810, -1,	-1,		-1,		"Hydrocast"},
		{820, -1,	-1,		-1,		"Thermistor Chain"},
		{830, -1,	-1,		-1,		"CTD"},
		{831, -1,	-1,		-1,		"CTD-P-ALACE Float"},
		{840, -1,	-1,		-1,		"PROVOR No Conductivity"},
		{841, -1,	-1,		-1,		"PROVOR Seabird Conductivity"},
		{842, -1,	-1,		-1,		"PROVOR FSI Conductivity"},
		{845, -1,	-1,		-1,		"Web Research No Conductivity"},
		{846, -1,	-1,		-1,		"Web Research Seabird Conductivity"},
		{847, -1,	-1,		-1,		"Web Research FSI Conductivity"},
		{850, -1,	-1,		-1,		"SOLO No Conductivity"},
		{851, -1,	-1,		-1,		"SOLO Seabird Conductivity"},
		{852, -1,	-1,		-1,		"SOLO FSI Conductivity"},
		{900, 9.727,	-0.0000473,		2000,	"Sippican T-12 (LMP5-T1)"}
	
	};
/*
struct Profiler

	{
		short InstrNo;
		double Acoeff;
		double Bcoeff;
		char *Instrument;

	};

*/
	for (int i = 0; i< sizeof Instr / sizeof Instr[0]; i++)
	{
		Instr[i].InstrNo = A[i].InstrNo;
		Instr[i].Acoeff = A[i].Acoeff;
		Instr[i].Bcoeff = A[i].Bcoeff;
		Instr[i].MaxDepth = A[i].MaxDepth;
		Instr[i].Instrument = A[i].Instrument;
	}
	Init(22067,1,0,0,10,1022,0,"INSTRUMENT TYPE FOR WATER PROFILE MEASUREMENT","CODE TABLE");
}

SingleCoeff FXY22067::GetCoeff()
{
	SingleCoeff A;
	A.InstrNo = -1;
	A.Acoeff = -1;
	A.Bcoeff = -1;
	for (int i=0;i < sizeof(Instr)/sizeof(Instr[0]);i++)
	{
		if(Instr[i].InstrNo==Value.Integ)
		{
			A.InstrNo = Instr[i].InstrNo;
			A.Acoeff = Instr[i].Acoeff;
			A.Bcoeff = Instr[i].Bcoeff;
			break;
		}
	}
	return A;
}

Profiler FXY22067::GetInstrument(int inst)
{	
	Profiler A;
	A.InstrNo = -1;
	A.Acoeff = -1;
	A.Bcoeff = -1;
	A.MaxDepth = -1;
	for (int i=0;i < sizeof(Instr)/sizeof(Instr[0]);i++)
	{
//		if(Instr[i].InstrNo==Value.Integ)
		if(Instr[i].InstrNo==inst)//JB I changed this from above value
		{
			A.InstrNo = Instr[i].InstrNo;
			A.Acoeff = Instr[i].Acoeff;
			A.Bcoeff = Instr[i].Bcoeff;
			A.MaxDepth = Instr[i].MaxDepth;
			A.Instrument = Instr[i].Instrument;
			break;
		}
	}
	return A;
}

int FXY22067::SetValue(int Inst)
{
	for (int i=0;i < sizeof(Instr)/sizeof(Instr[0]);i++)
	{
		if(Instr[i].InstrNo==Inst)
		{
			Value.Integ = Inst;
			Valid = true;
			return 0;
		}
	}

	Value.Integ = -1;
	Valid = false;
	return 0;
}	


// return the number of instruments in the instrument code table
int FXY22067::GetNoInst()
{
	int a = sizeof(Instr)/sizeof(Instr[0]);
	return a ;
}

// return the instrument at pos in a Profiler Struct
Profiler FXY22067::GetInstAt(int pos)
{
	if(pos > GetNoInst()-1)
	{
		Profiler A = {NULL, NULL,NULL,NULL};
		return A;
	}
	else
	{
		Profiler A;

		A.InstrNo = Instr[pos].InstrNo;
		A.Acoeff = Instr[pos].Acoeff ;
		A.Bcoeff = Instr[pos].Bcoeff;
		A.MaxDepth = Instr[pos].MaxDepth;
		A.Instrument = Instr[pos].Instrument;
		return A;
	}
}

//FXY22068    PROFILE MEASUREMENT RECORDER TYPE

FXY22068::FXY22068()
{

	
	CodeTable A[ Elem22068] = 
	{
		{1,"Sippican Strip Chart Recorder"},
		{2,"Sippican MK2A/SSQ-61"},
		{3,"Sippican MK-9"},
		{4,"Sippican AN/BHQ-7/MK8"},
		{5,"Sippican MK-12"},
		{6,"Sippican MK-21"},
		{10,"Sparton SOC BT/SV Processor Model 10"},
		{11,"Lockheed-Sanders Model QL5005"},
		{20,"ARGOS XBT-ST"},
		{21,"CLS-ARGOS/Protecno XBT-ST Model-1"},
		{22,"CLS-ARGOS/Protecno XBT-ST Model-2"},
		{30,"Bathy Systems SA-810"},
		{31,"Scripps Metrobyte Controller"},
		{32,"Murayama Denki Z-60-16 III"},
		{33,"Murayama Denki Z-60-16 II"},
		{34,"Protecno ETSM2"},
		{35,"Nautilus Marine Service NMS-XBT"},
		{40,"TSK MK-2A"},
		{41,"TSK MK-2S"},
		{42,"TSK MK-30"},
		{43,"TSK MK-30N"},
		{45,"TSK MK-100"},
		{46,"TSK MK-130"},
		{48,"TSK AXBT Receiver MK-300"},
		{50,"JMA ASTOS"},
		{60,"ARGOS Communications, Sampling on up transit"},
		{61,"ARGOS Communications, Sampling on down transit"},
		{62,"Orbcomm Communications, Sampling on up transit"},
		{63,"Orbcomm Communications, Sampling on down transit"},
		{70,"CSIRO Devil-1"},
		{71,"CSIRO Devil-2"},
		{99, "Unknown"}
	
	};

	for (int i = 0; i< Elem22068; i++)
	{
		Table.push_back(A[i]);
	}


	Init(22068,1,0,0,7,126,0,"PROFILE MEASUREMENT RECORDER TYPE","CODE TABLE");
}
// FXY22178 XBT/XCTD Launcher type
FXY22178::FXY22178()
{
	
	CodeTable A[ Elem22178] = 
	{
		{0, "Unknown"},
		{1, "LM-2A Deck-mounted"},
		{2, "LM-3A Hand-Held"},
		{3, "LM-4A Thru-Hull"},
		{10, "AL-12 TSK Autolauncher (up to 12 Probes)"},
		{20, "SIO XBT Autolauncher (up to 6 probes)"},
		{30, "AOML XBT V6 Autolauncher (up to 6 Deep Blue probes)"},
		{31, "AOML XBT V8.0 Autolauncher (up to 8 Deep Blue probes)"},
		{32, "AOML XBT V8.1 Autolauncher (up to 8 Deep Blue and Fast Deep probes)"},
		{90, "CSIRO Devil Autolauncher"},
		{91, "TURO/CSIRO Quoll Autolauncher"},
		{100, "MFSTEP Autolauncher (Mediterranean)"},
		{255, "Missing"}
	};
	for (int i = 0; i< Elem22178; i++)
	{
		Table.push_back(A[i]);
	}


	Init(22178,1,0,0,8,255,0,"XBT/XCTD LAUNCHER TYPE","CODE TABLE");
};

//FXY22195 HEIGHT OF WIND WAVES 
FXY22195::FXY22195()
{
	Init(22195,1,0,0,5,30,0,"HEIGHT OF WIND WAVES","HALF METERS");
}

//FXY22196 HEIGHT OF SWELL WAVES 
FXY22196::FXY22196()
{
	Init(22196,1,0,0,6,62,0,"HEIGHT OF SWELL WAVES","HALF METERS");
}

//FXY31001	NUMBER OF TIMES REPLICATED

FXY31001::FXY31001()
{
	Init(31001,1,0,0,4,14,0,"NUMBER OF GROUPS REPLICATED","NUMERIC");
}

//FXY31001	NUMBER OF TIMES REPLICATED

FXY31001A::FXY31001A()
{
	Init(31001,1,0,0,3,6,0,"NUMBER OF GROUPS REPLICATED","NUMERIC");
}

// FXY310001B used to indicate number of 5 character FXY205030A blocks make up 
// a text block
FXY31001B::FXY31001B()
{
	Init(31001,1,0,0,6,62,0,"NUMBER OF BLOCKS REPLICATED","NUMERIC");
}

//FXY31200	NUMBER OF TIMES REPLICATED

FXY31200::FXY31200()
{
	Init(31200,1,0,0,16,65534,0,"NUMBER OF TIMES REPLICATED","NUMERIC");
}

//FXY33003   Data Quality Indicator

FXY33003::FXY33003()
{
	CodeTable A[Elem33003] =
	{
		{0,"Data Not Suspect"},
		{1,"Data Slightly Suspect"},
		{2,"Data Highly Suspect"},
		{3,"Data Unfit For Use"},
	};
	for(int i =0; i< Elem33003; i++)
	{
		Table.push_back(A[i]);
	}


	Init(33003,1,0,0,3,6,0,"DATA QUALITY INFORMATION","CODE TABLE");
}


//FXY48001 

// Generic identfication flags used to determine whether or not an attribute is included in a message
// I.E. is the high resolution temperature transmitted in the TSG message.
// Use determined by the containing object
FXY48001::FXY48001()
{
	Init(48001,1,0,0,8,254,0,"VALUE INCLUDED IN MESSAGE FLAGS","NUMERIC");
}

//FXY48004 Ship's Direction
FXY48004::FXY48004()
{
	Init(48004,1,0,0,9,510,0,"DIRECTION SHIP IS HEADING","DEGREES TRUE");
}

//FXY48194 Ship's Speed
FXY48031::FXY48031()
{
	Init(48031,1,0,0,13,8190,2,"SHIP'S SPEED","METERS/SEC");
}

//FXY48194 AMVER PORT NUMBER
FXY48194::FXY48194()
{
	Init(48194,1,0,0,13,8190,0,"AMVER PORT NUMBER","NUMERIC");
}

//FXY48197 AMVER Navigation Method
FXY48197::FXY48197()
{
	Init(48197,1,0,0,2,2,0,"AMVER Navigation Method ","FLAG TABLE");
}

//FXY48197A AMVER Navigation Method
FXY48197A::FXY48197A()
{
	Init(48197,1,0,0,3,6,0,"AMVER Navigation Method ","FLAG TABLE");
}


//FXY48198 AMVER Medical Resources
FXY48198::FXY48198()
{
	Init(48198,1,0,0,4,14,0,"AMVER Medical Resources ","FLAG TABLE");
}


//FXY48199 AMVER DISTRIBUTION 
FXY48199::FXY48199()
{
	Init(48199,1,0,0,5,30,0,"AMVER DISTRIBUTION ","FLAG TABLE");
}
//FXY48200 RADIO STATION
FXY48200::FXY48200()
{
	Init(48200,1,0,0,24,0,0,"RADIO STATION","CCITT IA5");
}

//FXY48201 SEAS SOFTWARE VERSION NUMBER
FXY48201::FXY48201()
{
	Init(48201,1,0,0,10,999,0,"SEAS SOFTWARE VERSION NUMBER","NUMERIC");
}

//FXY48202 INMARSAT-C TRANSCEIVER MANUFACTURER CODE 
FXY48202::FXY48202()
{
	Init(48202,1,0,0,24,0,0,"INMARSAT-C TRANSCEIVER MANUFACTURER CODE","CCITT IA5");
}
//FXY48203 GPS MANUFACTURER CODE
FXY48203::FXY48203()
{
	Init(48203,1,0,0,24,0,0,"GPS MANUFACTURER CODE","CCITT IA5");
}

//FXY48204 Transmitter Software Version
FXY48204::FXY48204()
{
	Init(48204,1,0,2,10,1000,0,"GPS MANUFACTURER CODE","NUMERIC");
}

//FXY48207 INMARSAT OCEAN AREAS
FXY48207::FXY48207()
{
	Init(48207,1,0,0,3,4,0,"INMARSAT OCEAN AREAS","FLAG TABLE");
}


//FXY48208 INMARSAT-C MOBIL ID NUMBER FOR EMERGENCY

FXY48208::FXY48208()
{
	Init(48208,1,0,0,31,999999999,0,"INMARSAT-C MOBIL ID NUMBER FOR EMERGENCY","NUMERIC");
}


//FXY48211

FXY48211::FXY48211()
{
	Init(48211,1,0,0,6,62,0,"INMARSAT REPORT SUB-CATEGORY","CODE TABLE");
}


//FXY48212 INMARSAT-A MOBIL ID NUMBER

FXY48212::FXY48212()
{
	Init(48212,1,0,0,24,9999999,0,"INMARSAT-A MOBIL ID NUMBER (defunct)","NUMERIC");
}

//FXY48213 ADMIN REPORT TYPE
FXY48213::FXY48213()
{
	Init(48213,1,0,0,3,5,0,"ADMIN REPORT TYPE","FLAG TABLE");
}

//FXY48214 MAXIMUM SHIP SPEED KNOTS
FXY48214::FXY48214()
{
	Init(48214,1,0,1,9,500,0,"MAXIMUM SHIP SPEED KNOTS","NUMERIC");
}

//FXY48216  

FXY48216::FXY48216()
{
	Init(48216,1,0,0,10,1022,0,"SEAS2K INMARSAT REPORT SUB-CATEGORY","CODE TABLE");
}

//FXY48300  Probe Serial Number  

FXY48300::FXY48300()
{
	Init(48300,1,0,0,20,999999,0,"SEAS2K PROBE SERIAL NUMBER","NUMERIC");
}

// FXY48300A PROBE SERIAL NUMBER
FXY48300A::FXY48300A()
{
	Init(48300,1,0,0,24,16777214,0,"SEAS2K PROBE SERIAL NUMBER","NUMERIC");
}
//
//FXY48301  Cruise Name

FXY48301::FXY48301()
{
	Init(48301,1,0,0,160,999999,20,"CRUISE NAME","CCITT IA5");
}
// long
// FXY48302  Almost Unique Tag
FXY48302::FXY48302()
{
//	Init(48302,1,0,0,32,2147483646,0,"UNIQUE TAG","NUMERIC");
	Init(48302,1,0,0,32,(unsigned long) 4294967293,0,"UNIQUE TAG","NUMERIC");
}

// FXY48303  Type Data In Message
FXY48303::FXY48303()
{
	CodeTable A[Elem48303] =
	{
		{1, "Full Resolution"},
		{2, "2 Meter Resolution"},
		{3, "Inflection Points"}
	};
	for(int i =0; i< Elem48303; i++)
	{
		Table.push_back(A[i]);
	}
//	Table.push_back(CodeTable A = {2, "2 Meter Resolution"});
//	Table.push_back(CodeTable A = {1, "Inflection Points"});
/*	CodeTable A[sizeof Table/sizeof Table[0]] =
	{
		{1, "Full Resolution"},
		{2, "2 Meter Resolution"},
		{3, "Inflection Points"}
	};

	for (int i = 0; i< sizeof Table / sizeof Table[0]; i++)
	{
		Table[i].Code = A[i].Code;
		Table[i].Definition = A[i].Definition;
	}
*/

	Init(48303,1,0,0,3,6,0,"TYPE OF DATA","CODE TABLE");
}

/*CodeTable FXY48303::GetTableElem(int code)
{
	CodeTable A;
	A.Code = NULL;
	A.Definition = NULL;
	for (int i = 0; i< sizeof Table / sizeof Table[0]; i++)
	{
		if(Table[i].Code == code)
		{
			A.Code = Table[i].Code;
			A.Definition = Table[i].Definition;
		}
	}
	return A;
}

CodeTable FXY48303::GetTableElemAt(int pos)
{
	CodeTable A;
	A.Code = NULL;
	A.Definition = NULL;
	if((pos >=0)&& (pos < GetNoElems()))
	{
		A.Code = Table[pos].Code;
		A.Definition = Table[pos].Definition;
	}

	return A;
}

int FXY48303::GetNoElems()
{
	int a = sizeof(Table)/sizeof(Table[0]);
	return a ;
}


*/

// Message Sequence Number
FXY48304::FXY48304()
{
	Init(48304,1,0,0,16,65534,1,"Message Sequence Number","NUMERIC");
}

// XBT Launch Height
FXY48305::FXY48305()
{
	Init(48305,1,0,0,13,8190,2,"Launch Height","METERS");
}
// FXY48604  Averaging period for salinity, Sea temperature and surface current speed/direction
FXY48604::FXY48604()
{
/*	CodeTable A[Elem48604] =
	{
		{0, "Spot Values"},
		{1, "Less than 15 minutes"},
		{2, "From 15 to 45 minutes"},
		{3, "More than 45 minutes"},
		{9, "Data not available"}
	};*/

	A[0].Code = 0;
	A[0].Definition ="Spot Values";
	A[1].Code=1;
	A[1].Definition ="Less than 15 minutes";
	A[2].Code=2;
	A[2].Definition ="From 15 to 45 minutes";
	A[3].Code=3;
	A[3].Definition ="More than 45 minutes";
	A[4].Code=9;
	A[4].Definition ="Data not available";


	for(int i =0; i< Elem48604; i++)
	{
		Table.push_back(A[i]);
	}
//	Table.push_back(CodeTable A = {2, "2 Meter Resolution"});
//	Table.push_back(CodeTable A = {1, "Inflection Points"});
/*	CodeTable A[sizeof Table/sizeof Table[0]] =
	{
		{1, "Full Resolution"},
		{2, "2 Meter Resolution"},
		{3, "Inflection Points"}
	};

	for (int i = 0; i< sizeof Table / sizeof Table[0]; i++)
	{
		Table[i].Code = A[i].Code;
		Table[i].Definition = A[i].Definition;
	}
*/


	Init(48604,1,0,0,3,6,0,"AVERAGING PERIOD","CODE TABLE");
}

//
int FXY48604::SetValue(int val)
{
	for (int i=0;i < Elem48604;i++)
	{
		if(A[i].Code==val)
		{
			Value.Integ = val;
			Valid = true;
			return 0;
		}
	}

	Value.Integ = -1;
	Valid = false;
	return 0;
}	


//
//FXY48301  Cruise Name

//FXY100000  Delayed Replicator!
// Different from BUFR.  The number contained within is the number of following 
// consecutive fields to replicate.  This field must be followed by FXY31200 to identify
// how many groups of replicated fields follow.  Always delayed replication

FXY100000::FXY100000()
{
	Init(100000,1,0,0,8,254,0,"SEAS2K Replication factor","NUMERIC");
}

FXY104000::FXY104000()
{
	Init(104000,1,0,0,4,14,0,"SEAS2K Replication factor","NUMERIC");
}

// FXY205030 COMMENTS
FXY205030::FXY205030()
{
	Init(205030,1,0,0,240,0,0,"COMMENTS","CCITT IA5");
}

// FXY205031 COMMENTS
FXY205030A::FXY205030A()
{
	Init(205030,1,0,0,40,0,0,"SHORT COMMENTS (Five Characters)","CCITT IA5");
}


//Class for Form 94 Elements with Code tables
//vector
CodeTableElem::CodeTableElem()
{}

CodeTable CodeTableElem::GetTableElem(int code)
{


	CodeTable A;
	A.Code = NULL;
	A.Definition = NULL;
	bool FallThru = true;
	for (int i = 0; i< (int)Table.size(); i++)
	{
		if(Table.at(i).Code == code)
		{
			A.Code = Table.at(i).Code;
			A.Definition = Table.at(i).Definition;
			FallThru = false;
			break;
		}
	}
	if(FallThru)
	{
		A.Code=-1;
		A.Definition="";
	}
	return A;
}

CodeTable CodeTableElem::GetTableElemAt(int pos)
{
	CodeTable A;
	A.Code = NULL;
	A.Definition = NULL;
	if((pos >=0)&& (pos < (int)Table.size()))
	{
		A.Code = Table.at(pos).Code;
		A.Definition = Table.at(pos).Definition;
	}
	else
	{
		A.Code=-1;
		A.Definition = "";
	}
	return A;
}

int CodeTableElem::GetNoElems()
{
	int a = (int)Table.size();
	return a ;
}
bool CodeTableElem::ValidValue(int code)
{	
	bool FallThru = true;
	for (int i = 0; i< (int)Table.size(); i++)
	{
		if(Table.at(i).Code == code)
		{
			FallThru = false;
			break;
		}
	}
	return  !FallThru;
}

//Class for Form 94 Agency Elements with Code tables  10/14/2014 PIC
//vector
AgencyTableElem::AgencyTableElem()
{}

AgencyTable AgencyTableElem::GetTableElem(int code)
{


	AgencyTable A;
	A.Code = NULL;
	A.Definition = NULL;
	bool FallThru = true;
	for (int i = 0; i< (int)Table.size(); i++)
	{
		if(Table.at(i).Code == code)
		{
			A.Code = Table.at(i).Code;
			A.Definition = Table.at(i).Definition;
			FallThru = false;
			break;
		}
	}
	if(FallThru)
	{
		A.Code=-1;
		A.Definition="";
	}
	return A;
}

AgencyTable AgencyTableElem::GetTableElemAt(int pos)
{
	AgencyTable A;
	A.Code = NULL;
	A.Definition = NULL;
	if((pos >=0)&& (pos < (int)Table.size()))
	{
		A.Code = Table.at(pos).Code;
		A.Definition = Table.at(pos).Definition;
	}
	else
	{
		A.Code=-1;
		A.Definition = "";
	}
	return A;
}

int AgencyTableElem::GetNoElems()
{
	int a = (int)Table.size();
	return a ;
}
bool AgencyTableElem::ValidValue(int code)
{	
	bool FallThru = true;
	for (int i = 0; i< (int)Table.size(); i++)
	{
		if(Table.at(i).Code == code)
		{
			FallThru = false;
			break;
		}
	}
	return  !FallThru;
}
