#include <jjvv_1.h>
#include <jjvv_2.h>
#include <math.h>

using namespace std;

/* PIC 09/20/11 modified to override DecodeElements and DecodeBinMess in JJVV. This allows
	a JJVV_1 type message to be decoded by a JJVV_2 object. If the UniqueTag is recomputed
	after a JJVV_1 has been decoded, the UniqueTag (CRC) will not equal the one embedded in the 
	JJVV_1 binary message.  This is due to the additional bits added in the JJVV_2 message
*/

//  * PIC 11/5/2014

JJVV_2::JJVV_2()
{
	DecodeType1 = false;
	SetMessType(2);
//#ifndef _WIN32
	int i;
//#endif

	for ( i =0; i< SEAS_Elem.size();i++)
	{
		FM94Elem *Val;
		int AT = -1;
		Val = (FM94Elem *)SEAS_Elem.at(i);
		string a = Val->GetDesc();
		size_t kkk = a.find("PROBE SERIAL NUMBER");
		if(a.find("PROBE SERIAL NUMBER")!= (size_t) -1)
		{
			SEAS_Elem.erase(SEAS_Elem.begin()+i);
			SEAS_Elem.insert(SEAS_Elem.begin()+i,&(this->ProbeSerial));
			break;
		}
//		printf("%s\n",a.c_str());
	}
//	printf("\n");

	SEAS_Elem.insert(SEAS_Elem.begin()+17,&ShipSpeed);
	SEAS_Elem.insert(SEAS_Elem.begin()+17,&ShipDir);
	SEAS_Elem.insert(SEAS_Elem.begin()+17,&LaunchHeight);
	SEAS_Elem.insert(SEAS_Elem.begin()+5, &SeqNo);
	SEAS_Elem.insert(SEAS_Elem.begin()+5, &TransectNo);
	SEAS_Elem.insert(SEAS_Elem.begin()+5, &SOOPLine);
	ComputeMessSize();

	for (  i =0; i< SEAS_Elem.size();i++)
	{
		FM94Elem *Val;

		Val = (FM94Elem *)SEAS_Elem.at(i);
		string a = Val->GetDesc();
//		printf("%s\n",a.c_str());
	}
//	printf("\n");

	};


JJVV_2::~JJVV_2()
{};


void JJVV_2::SetShipSpeed(float SS) //Speed in Knots
{
//	float k= SS * pow(10,(double)  ShipSpeed.GetScale());
	float k = (SS *(1852.0/(3600.0*pow(10,(double) - ShipSpeed.GetScale()))));//+.5);
	int x = (int) (k+.5);
	ShipSpeed.SetValue(x);
}

void JJVV_2::SetShipDir(int SD)
{
	ShipDir.SetValue(SD);
};

void JJVV_2::SetTransectNo(int trans)
{
	TransectNo.SetValue(trans);
};

void JJVV_2::SetSeqNo(int SN)
{
	SeqNo.SetValue(SN);
};

void JJVV_2::SetSOOPLine(char *SOOP)
{
	SOOPLine.SetValue(SOOP);

};


void JJVV_2::SetLaunchHeight(float hgt) //Height in Meters to 2 decimal places
{
	float k = (hgt * pow(10.0,(double) LaunchHeight.GetScale()));//+.5);
	int x = (int) (k+.5);
	LaunchHeight.SetValue(x);
}

float JJVV_2::GetLaunchHeight() //Height in Meters to 2 decimal places
{
	if(LaunchHeight.GetValue().Integ==-1)
	{
		return -1;
	}

	float k = ((float) LaunchHeight.GetValue().Integ) * pow(10.0, -(double) LaunchHeight.GetScale());//+.5);
	return k;
}

float JJVV_2::GetShipSpeed() //Speed in Knots
{
	if(ShipSpeed.GetValue().Integ==-1)
	{
		return -1;
	}
	 float k = ShipSpeed.GetValue().Integ;
     float  x = (ShipSpeed.GetValue().Integ)/(1852.0/(3600.0*pow(10,(double) -ShipSpeed.GetScale())));
	 x= (float)((int)(x*pow(10.0,(double) ShipSpeed.GetScale())+.5));
	 x= x/pow(10.0,(double) ShipSpeed.GetScale());

	return x;
 }

int JJVV_2::GetShipDir()
{
	return ShipDir.GetValue().Integ;
}

int JJVV_2::GetSeqNo()
{
	return SeqNo.GetValue().Integ;
}
int JJVV_2::GetTransectNo()
{
	return TransectNo.GetValue().Integ;
}

string JJVV_2::GetSOOPLine()
{
//	string x = SOOPLine.GetValue().StrVal;
	string x = SOOPLine.GetBufrVal();
	if(SOOPLine.GetValue().Integ == 0xFFFFFFFF)
	{
		x="Missing";
	}
	return x;
}

std::string JJVV_2::NODCform( unsigned long *CRC)
{

	string Line1="";
	Line1 += NODCformMETA_1 (&*CRC);
	Line1 += NODCformMETA_2 ();
	Line1 += NODCformProfile();
	return Line1;
}

std::string JJVV_2::NODCformMETA_2()
{
	string Line, Line1;
	Line1="";
	Line = format(" Ship Speed at Launch (knots): %5.2f\r\n", GetShipSpeed());
	Line1 += Line;
	Line = format(" Ship Direction at Launch (Degrees): %4d\r\n", GetShipDir());
	Line1 += Line;
	Line = format(" Sequence Number: %5d\r\n",GetSeqNo());
	Line1 += Line;
	Line = format(" Transect Number: %4d\r\n", GetTransectNo());
	Line1 += Line;
	Line = format(" Launch Height (Meters): %5.2f\r\n", GetLaunchHeight());
	Line1 += Line;
	string soopline = GetSOOPLine();
	Line = format(" SOOP Line: %s\r\n",soopline.c_str());
	Line1 += Line;
	return Line1;
}
bool JJVV_2::DecodeElements(int start, int No_To_Decode, RetMess *InMess, int *BitPos)
{ 
	FM94Elem *y;
	FM94Value z;
	int size;
	int i;

//	printf("Start, No to Decode, bitpos %d, %d, %d\n",start, No_To_Decode, *BitPos);
	for ( i=start; i< start + No_To_Decode; i++)
	{
/*		if(i==17)
		{
			printf("I = 17");
		}
*/
		y = (FM94Elem *) SEAS_Elem.at(i);
		if(ExcludeElement(i))
		{
			string a = y->GetDesc();
//			printf("Excluding at %d  %s\n",i, a.c_str());
			continue;
		}

		// Just check to see if we may go outside message length
		// This theoretically should never happen
		size = y->GetMaxBits();
//		if((size+BitPos)/8 > InMess->size*8)
		if((size+*BitPos)/8 > InMess->size)
		{
			return false;
		}
		int j;
		for (j=0; j<sizeof(z.StrVal); j++)
		{
			z.StrVal[j] = (char) 255;
		}

		for (j=1; j<y->GetMaxBits()+1; j++)
		{
			if(CheckBit( *BitPos+j, InMess->Mess))
			{
				SetBit(j,1,z.StrVal);
			}
			else
			{
				SetBit(j,0,z.StrVal);
			}
		}

		int pp = strcmp(y->GetUnits(), "CCITT IA5");
		if(strcmp(y->GetUnits(), "CCITT IA5") != 0)
		{
			if(z.Integ!=-1)
			{

//				unsigned z1 = z.Integ;
				 // Swap the bytes to make integer appear as bit stream
				int noshifts = 32 - y->GetMaxBits();
				string xx = y->GetDesc();
				 char a = z.StrVal[0];
				 z.StrVal[0] = z.StrVal[3];
				 z.StrVal[3] = a;
				 a = z.StrVal[1];
				 z.StrVal[1] = z.StrVal[2];
				 z.StrVal[2] = a; 
				z.Integ = z.Integ >> noshifts;
				unsigned long mask = 0xffffffff;//(int) (pow((double) 2, (double) y->GetMaxBits() )-1.0);
				mask = mask >> noshifts;
				z.Integ=z.Integ & (unsigned) mask;				   

				y->SetValue((unsigned long) z.Integ);				
			}
		}
		else
		{
			y->SetValue(z.StrVal);
		}

		*BitPos += y->GetMaxBits();
	}
	return true;
}

bool JJVV_2::ExcludeElement(int Elno)
{
	// If type 1 decode, exclude the following Elements
	unsigned char Exclude[6]={5,6,7,20,21,22};
	int i;
	for(i=0;i<6;i++)
	{
		if((Elno == Exclude[i])&&(GetMessType()==1))
		{
			return true;
		}
	}
	return false;
}

bool JJVV_2::DecodeBinMess(RetMess *InMess)
{
	FM94Elem *y;
//	FM94Value z;
	FXY7062S Dpt;
	FXY22042S tmp, tmp1;
	DepTemp inf;
//	int size;
	int i;



//	SEAS_Elem.erase(SEAS_Elem.begin() +No_SEASOBS_Elements ,SEAS_Elem.end());

		int iii = SEAS_Elem.size();


/*	SEAS_Elem.push_back(&ShipName);
	SEAS_Elem.push_back(&Lloyds);
//	FXY48301 CruiseName;
	SEAS_Elem.push_back(&UniqueTag);
	SEAS_Elem.push_back(&SEASVersion);
	SEAS_Elem.push_back(&ProbeSerial);
	SEAS_Elem.push_back(&ThisDataIs);
	SEAS_Elem.push_back(&DataQual);
	SEAS_Elem.push_back(&InstType);
	SEAS_Elem.push_back(&RecorderType);
	SEAS_Elem.push_back(&TypeWindInst);
	SEAS_Elem.push_back(&WindDir);
	SEAS_Elem.push_back(&WindSpeed);
	SEAS_Elem.push_back(&DryBulbTemp);
	SEAS_Elem.push_back(&CurrMeasMeth);
	SEAS_Elem.push_back(&CurrentDir);
	SEAS_Elem.push_back(&CurrentSpeed);
	SEAS_Elem.push_back(&TotalWaterDepth);
	SEAS_Elem.push_back(&NoRepFields);
	SEAS_Elem.push_back(&TimesReplicated);
//	SEAS_Elem.push_back(&SeaTemp);
//	SEAS_Elem.push_back(&SeaDepth);
*/
	int BitPos = 0;

	if(	BinaryMess != NULL)
	{
		delete [] BinaryMess;
		BinaryMess = NULL;
	}

	TailsRem = false;
	OneMeterComputed = false;
	TailDepth = -1;
	InflectComp = false;

	Inflection.clear();
	MySeaTemp.clear();
	OneMeter.clear();
	RawSeaTemp.clear();

	//  Decode the General SEASOB Information..


	int Elements_Now = SEAS_Elem.size();


// Decode all of the general SEASOB Meta Info

	if(!DecodeElements(0,No_SEASOBS_Elements, InMess, &BitPos))
	{
		return false;
	}

	if(GetMessType()==1)
	{
		DecodeType1=true;
		SEAS_Elem.erase(SEAS_Elem.begin()+17);
		SEAS_Elem.insert(SEAS_Elem.begin()+17,&(JJVV::ProbeSerial));
		ComputeMessSize();
	}

// Decode the JJVV Meta Info

	if(!DecodeElements(No_SEASOBS_Elements,Elements_Now-No_SEASOBS_Elements, InMess, &BitPos))
	{
		if(DecodeType1)
		{
			ResetProbeOnVect();
		}
		return false;
	}

	if(ThisDataIs.GetValue().Integ == 1)  // Full resolution profile being decoded
	{
		int test = TimesReplicated.GetValue().Integ;
		for(i=0; i< TimesReplicated.GetValue().Integ;i++)
		{
			if(!JJVV::DecodeElements(InMess, &BitPos, &tmp))
			{
				if(DecodeType1)
				{
					ResetProbeOnVect();
				}
				return false;
			}
			else
			{
				MySeaTemp.push_back(tmp);
				float x=tmp.GetValue().Integ;
				x = (x+tmp.GetOffset())/pow(10.0, tmp.GetScale());
				RawSeaTemp.push_back(x);
			}

		}
	}
		/* two meter data being decode therefore populate onemeter vector with two copies
		of the two meter element.  This provide compatibility of other routins that use the onemeter vector.
		We are basically using an even two meter value to also represent the previous odd value
		*/
	else if(ThisDataIs.GetValue().Integ == 2) 
	{
		for(i=0; i< TimesReplicated.GetValue().Integ;i++)
		{
			if(!JJVV::DecodeElements(InMess, &BitPos, &tmp))
			{
				if(DecodeType1)
				{
					ResetProbeOnVect();
				}
				return false;
			}
			else
			{
				OneMeterComputed=true;
				OneMeter.push_back(tmp);
				OneMeter.push_back(tmp);
			}
		}
	}
	else if(ThisDataIs.GetValue().Integ == 3)
	{
		for(i=0; i< TimesReplicated.GetValue().Integ;i++)
		{
			for(int k=0; k<2; k++)
			{
				if(k==0)
				{
					y = (FM94Elem *) &Dpt;
				}
				else
				{
					y = (FM94Elem *) &tmp;
				}

				if(!JJVV::DecodeElements(InMess, &BitPos, y))
					{
						if(DecodeType1)
						{
							ResetProbeOnVect();
						}
						return false;
					}
			}
					 
			inf.Depth = Dpt.GetValue().Integ;
			float x=tmp.GetValue().Integ;
			x = (x+tmp.GetOffset())/pow(10.0, tmp.GetScale());
			inf.Temp = x;
			Inflection.push_back(inf);
		}
	}

//pic	JJVV::ProbeSerial.SetValue(34567);
	if(DecodeType1)
	{
		ResetProbeOnVect();
	}

	return true;
}
int  JJVV_2::GetProbeSerial()//Added JB
{	
	return 	ProbeSerial.GetValue().Integ;

}

void JJVV_2::SetProbeSerial(int SN)
{
	ProbeSerial.SetValue(SN);
};

void JJVV_2::ResetProbeOnVect()
{
		// We decoded a type 1 message.  Set the message type to 2 
		// and copy the 20 bit probe serial number to the 24 bit probe serial number.
		// replace the ProbeSerial on the vector and recompute the message size

		SetMessType(2);
		DecodeType1=false;
		int ps = JJVV::GetProbeSerial();
		this->ProbeSerial.SetValue(JJVV::GetProbeSerial());
		SEAS_Elem.erase(SEAS_Elem.begin()+17);
		SEAS_Elem.insert(SEAS_Elem.begin()+17,&(this->ProbeSerial));
		ComputeMessSize();
}
